using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Windows.Forms.Layout;
using LinearProgramming;

namespace ALM
{
    /// <summary>
    /// The possibilities for adjusting a GUI's layout.
    /// Either change the child controls so that they fit into their parent control,
    /// or adjust the size of the parent control so that the children have as much space as they want.
    /// </summary>
    public enum LayoutStyleType { FIT_TO_SIZE, ADJUST_SIZE };

    /// <summary>
    /// A GUI layout engine using the ALM.
    /// </summary>
    public class ALMEngine : LayoutEngine
    {
        /// <summary>
        /// The specification used for calculating the layout.
        /// </summary>
        public LayoutSpec LayoutSpec;

        /// <summary>
        /// The manner in which the GUI is dynamically adjusted.
        /// The default is to fit the child controls into their parent.
        /// </summary>
        public LayoutStyleType LayoutStyle = LayoutStyleType.FIT_TO_SIZE;
        bool activated = true;
        
        //Parent control
        internal Control parent;
        

        /// <summary>
        /// Creates a new layout engine with an empty specification.
        /// </summary>
        public ALMEngine() : base() { }

        /// <summary>
        /// Creates a new layout engine with the given specification.
        /// </summary>
        /// <param name="ls">the layout specifiaction</param>
        public ALMEngine(LayoutSpec ls)
            : base()
        {
            LayoutSpec = ls;
        }

        public override bool Layout(object container, LayoutEventArgs layoutEventArgs)
        {
            parent = container as Control;
            Layout();
            return false;  // False means that container's parent should not perform layout.
        }

        /// <summary>
        /// Calculate and set the layout.
        /// If no layout specification is given, a specification is reverse engineered automatically.
        /// </summary>
        /// <param name="parent">the parent control of the controls in the layout</param>
        public void Layout()
        {
            // make sure that layout events occuring during layout are ignored
            // i.e. activated is set to false during layout caluclation
            if (!activated) return;
            activated = false;

            // reverse engineer a layout specification if none was given
            if (LayoutSpec == null) RecoverLayout(parent);

            // if the layout engine is set to fit the GUI to the given size,
            // then the given size is enforced by setting absolute positions for Right and Bottom
            if (LayoutStyle == LayoutStyleType.FIT_TO_SIZE)
            {
                LayoutSpec.Right.SetRange(parent.DisplayRectangle.Width, parent.DisplayRectangle.Width);
                LayoutSpec.Bottom.SetRange(parent.DisplayRectangle.Height, parent.DisplayRectangle.Height);
            }

            LayoutSpec.SolveLayout();

            // if new layout is infasible, use previous layout
            if (LayoutSpec.Result == ResultType.INFEASIBLE)
            {
                activated = true; // now layout calculation is allowed to run again
                return;
            }
            if (LayoutSpec.Result != ResultType.OPTIMAL)
            {
                LayoutSpec.Save("failed-layout.txt");
                throw new Exception("Could not solve the layout specification ("
                    + LayoutSpec.Result.ToString() + "). Saved specification in file failed-layout.txt");
            }

            //LayoutSpec.Save("last-feasible-layout.txt");

            // change the size of the GUI according to the calculated size 
            // if the layout engine was configured to do so
            if (LayoutStyle == LayoutStyleType.ADJUST_SIZE)
            {
                parent.Size = new Size(
                    (int)Math.Round(LayoutSpec.Right.Value - LayoutSpec.Left.Value),
                    (int)Math.Round(LayoutSpec.Bottom.Value - LayoutSpec.Top.Value));
            }

            // set the calculated positions and sizes for every area
            foreach (Area a in LayoutSpec.Areas) a.DoLayout(editing);

            activated = true; // now layout calculation is allowed to run again
        }

        #region Editing
        /// <summary>
        /// Indicates whether editing mode is on.
        /// </summary>
        bool editing = false;

        /// <summary>
        /// List that stores original controls while in edit mode
        /// </summary>
        List<Control> savedControls;

        /// <summary>
        /// Edit form used
        /// </summary>
        public EditForm editForm;

        public void Edit()
        {

            activated = false;
            editing = true;
            this.LayoutSpec.Solve();
            this.editForm = new EditForm(parent, this);
            savedControls = new List<Control>();
            foreach (Control control in parent.Controls)
            {
                savedControls.Add(control);
            }
            activated = true;
            editForm.Show();
            editForm.UpdateEditForm();
        }

        public void QuitEdit()
        {
            editing = false;
            parent.Controls.Clear();
            foreach (Control control in savedControls)
            {
                parent.Controls.Add(control);
            }
        }
        #endregion

        #region RE
        /// <summary>
        /// Reverse engineers a GUI and recovers an ALM specification.
        /// </summary>
        /// <param name="parent">the parent container of the GUI</param>
        public void RecoverLayout(Control parent)
        {
            LayoutSpec = new LayoutSpec();
            SortedList<int, XTab> xtabs = new SortedList<int, XTab>();
            SortedList<int, YTab> ytabs = new SortedList<int, YTab>();

            xtabs.Add(parent.DisplayRectangle.Left, LayoutSpec.Left);
            xtabs.Add(parent.DisplayRectangle.Right, LayoutSpec.Right);
            ytabs.Add(parent.DisplayRectangle.Top, LayoutSpec.Top);
            ytabs.Add(parent.DisplayRectangle.Bottom, LayoutSpec.Bottom);

            foreach (Control c in parent.Controls)
            {
                XTab x1, x2;
                YTab y1, y2;

                if (xtabs.ContainsKey(c.Left)) x1 = xtabs[c.Left];
                else xtabs.Add(c.Left, x1 = new XTab(LayoutSpec));
                if (xtabs.ContainsKey(c.Right)) x2 = xtabs[c.Right];
                else xtabs.Add(c.Right, x2 = new XTab(LayoutSpec));
                if (ytabs.ContainsKey(c.Top)) y1 = ytabs[c.Top];
                else ytabs.Add(c.Top, y1 = new YTab(LayoutSpec));
                if (ytabs.ContainsKey(c.Bottom)) y2 = ytabs[c.Bottom];
                else ytabs.Add(c.Bottom, y2 = new YTab(LayoutSpec));

                LayoutSpec.AddArea(x1, y1, x2, y2, c);

                x2.LeftLink = true;
                y2.TopLink = true;
            }

            // adding additional constraints (links in the PO) for margins between areas
            for (int i = 1; i < xtabs.Count; i++)
            {
                if (!xtabs.Values[i].LeftLink)
                    LayoutSpec.AddConstraint(-1, xtabs.Values[i - 1], 1, xtabs.Values[i],
                        OperatorType.EQ, -xtabs.Keys[i - 1] + xtabs.Keys[i]);
            }
            for (int i = 1; i < ytabs.Count; i++)
            {
                if (!ytabs.Values[i].TopLink)
                    LayoutSpec.AddConstraint(-1, ytabs.Values[i - 1], 1, ytabs.Values[i],
                        OperatorType.EQ, -ytabs.Keys[i - 1] + ytabs.Keys[i]);
            }
        }
        #endregion RE
    }
}