using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using LinearProgramming;

namespace ALM
{
    /// <summary>
    /// The options available for horizontal alignment.
    /// </summary>
    public enum HorizontalAlignment { LEFT, RIGHT, CENTER, FILL, NONE };

    /// <summary>
    /// The options available for vertical alignment.
    /// </summary>
    public enum VerticalAlignment { TOP, BOTTOM, CENTER, FILL, NONE };

    /// <summary>
    /// Rectangular area in the GUI, defined by a tab on each side.
    /// </summary>
    public class Area
    {
        /// <summary>
        /// Minimum possible size. Use this if there is no lower bound.
        /// </summary>
        public static Size MIN_SIZE = new Size(0, 0);

        /// <summary>
        /// Maximum possible size. Use this if there is no upper bound.
        /// </summary>
        public static Size MAX_SIZE = new Size(Int32.MaxValue, Int32.MaxValue);

        /// <summary>
        /// Undefined size. Used if a certain size constraint is not set.
        /// </summary>
        public static Size UNDEFINED_SIZE = new Size(-1, -1);

        /// <summary>
        /// The layout specification this area belongs to.
        /// </summary>
        LayoutSpec ls;

        XTab left, right;
        YTab top, bottom;
        Row row;
        Column column;
        internal Control content;
        /// <summary>
        /// A list of constraints which are removed form the speciifcation when the area is removed.
        /// </summary>
        internal List<Constraint> Constraints = new List<Constraint>();

        Size minContentSize = new Size(0, 0);
        Size maxContentSize = MAX_SIZE;

        /// <summary>
        /// Size constraint for the content. Valid even if the content is actually in a child area.
        /// </summary>
        Constraint minContentWidth, maxContentWidth, minContentHeight, maxContentHeight;

        Size prefContentSize = UNDEFINED_SIZE;
        SizeF shrinkRigidity = new SizeF(2, 2);
        SizeF expandRigidity = new SizeF(1, 1);

        double contentAspectRatio = Double.NaN;
        Constraint contentAspectRatioC;

        /// <summary>
        /// Determines if the PrefContentSize is set automatically.
        /// Manual changes of PrefContentSize are ignored unless set to false.
        /// </summary>
        public bool AutoPrefContentSize = false;

        /// <summary>
        /// The soft constraints used for setting a preferred width.
        /// </summary>
        Constraint prefContentWidth;

        /// <summary>
        /// The soft constraints used for setting a preferred height.
        /// </summary>
        Constraint prefContentHeight;

        /// <summary>
        /// This area is used if the borders of the control are not the same as the borders of the area,
        /// i.e. if the control is smaller than the area due to margins or alignment.
        /// In case a childArea is present, many of the operations (such as those for setting content sizes)
        /// are delegated to the childArea.
        /// </summary>
        Area childArea;

        HorizontalAlignment hAlignment = HorizontalAlignment.FILL;
        VerticalAlignment vAlignment = VerticalAlignment.FILL;
        int leftMargin = 0;
        int topMargin = 0;
        int rightMargin = 0;
        int bottomMargin = 0;

        /// <summary>
        /// The constraint for setting margin and alignment on the left.
        /// </summary>
        Constraint leftConstraint;

        /// <summary>
        /// The constraint for setting margin and alignment at the top.
        /// </summary>
        Constraint topConstraint;

        /// <summary>
        /// The constraint for setting margin and alignment on the right.
        /// </summary>
        Constraint rightConstraint;

        /// <summary>
        /// The constraint for setting margin and alignment at the bottom.
        /// </summary>
        Constraint bottomConstraint;

        /// <summary>
        /// Left tab of the area.
        /// </summary>
        public XTab Left
        {
            get { return left; }
            set
            {
                left = value;
                column = null; // since we changed an individual x-tab we do not align to a Column anymore

                if (childArea == null)
                {
                    // the respective minimum constraint needs to use the new tab
                    minContentWidth.SetLeftSide(-1, left, 1, right);

                    // if a maximum constraint was set, then it nees to use the new tab
                    if (maxContentWidth != null)
                        maxContentWidth.SetLeftSide(-1, left, 1, right);
                }
                else
                    updateHorizontal(); // the constraints between outer area and childArea need to use the new tab
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Right tab of the area.
        /// </summary>
        public XTab Right
        {
            get { return right; }
            set
            {
                right = value;
                column = null; // since we changed an individual x-tab we do not align to a Column anymore

                if (childArea == null)
                {
                    // the respective minimum constraint needs to use the new tab
                    minContentWidth.SetLeftSide(-1, left, 1, right);

                    // if a maximum constraint was set, then it nees to use the new tab
                    if (maxContentWidth != null)
                        maxContentWidth.SetLeftSide(-1, left, 1, right);
                }
                else
                    updateHorizontal(); // the constraints between outer area and childArea need to use the new tab
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Top tab of the area.
        /// </summary>
        public YTab Top
        {
            get { return top; }
            set
            {
                top = value;
                row = null; // since we changed an individual y-tab we do not align to a Row anymore

                if (childArea == null)
                {
                    // the respective minimum constraint needs to use the new tab
                    minContentHeight.SetLeftSide(-1, top, 1, bottom);

                    // if a maximum constraint was set, then it nees to use the new tab
                    if (maxContentHeight != null)
                        maxContentHeight.SetLeftSide(-1, top, 1, bottom);
                }
                else
                    updateVertical(); // the constraints between outer area and childArea need to use the new tab
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Bottom tab of the area.
        /// </summary>
        public YTab Bottom
        {
            get { return bottom; }
            set
            {
                bottom = value;
                row = null; // since we changed an individual y-tab we do not align to a Row anymore

                if (childArea == null)
                {
                    // the respective minimum constraint needs to use the new tab
                    minContentHeight.SetLeftSide(-1, top, 1, bottom);

                    // if a maximum constraint was set, then it nees to use the new tab
                    if (maxContentHeight != null)
                        maxContentHeight.SetLeftSide(-1, top, 1, bottom);
                }
                else
                    updateVertical(); // the constraints between outer area and childArea need to use the new tab
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// The row that defines the top and bottom tabs. May be null.
        /// </summary>
        public Row Row
        { 
            get { return row; }
            set
            {
                Top = value.Top;
                Bottom = value.Bottom;
                row = value;
                ls.InvalidateLayout();
            }
        }
        
        /// <summary>
        /// The column that defines the left and right tabs. May be null.
        /// </summary>
        public Column Column 
        { 
            get { return column; }
            set
            {
                Left = value.Left;
                Right = value.Right;
                column = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// The control that is the content of the area.
        /// </summary>
        public Control Content
        {
            get {
                // if we are using a childArea then it contains the control,
                // otherwise this area contains the control directly
                return (childArea == null) ? content : childArea.Content; 
            }
            set
            {
                if (childArea == null) content = value;
                else childArea.content = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Left tab of the area's content. May be different from the left tab of the area.
        /// </summary>
        public XTab ContentLeft
        {
            get { return (childArea == null) ? left : childArea.left; }
        }

        /// <summary>
        /// Top tab of the area's content. May be different from the top tab of the area.
        /// </summary>
        public YTab ContentTop
        {
            get { return (childArea == null) ? top : childArea.top; }
        }

        /// <summary>
        /// Right tab of the area's content. May be different from the right tab of the area.
        /// </summary>
        public XTab ContentRight
        {
            get { return (childArea == null) ? right : childArea.right; }
        }

        /// <summary>
        /// Bottom tab of the area's content. May be different from the bottom tab of the area.
        /// </summary>
        public YTab ContentBottom
        {
            get { return (childArea == null) ? bottom : childArea.bottom; }
        }

        /// <summary>
        /// Minimum size of the area's content. May be different from the minimum size of the area.
        /// </summary>
        public Size MinContentSize
        {
            get { return (childArea == null) ? minContentSize : childArea.minContentSize; }
            set
            {
                if (childArea == null)
                {
                    minContentSize = value;
                    minContentWidth.RightSide = minContentSize.Width;
                    minContentHeight.RightSide = minContentSize.Height;
                }
                else childArea.MinContentSize = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Maximum size of the area's content. May be different from the maximum size of the area.
        /// </summary>
        public Size MaxContentSize
        {
            get { return (childArea == null) ? maxContentSize : childArea.maxContentSize; }
            set
            {
                if (childArea == null)
                {
                    maxContentSize = value;
                    if (maxContentWidth == null) // no max constraints set yet
                    {
                        maxContentWidth = ls.AddConstraint(-1, left, 1, right, OperatorType.LE, maxContentSize.Width);
                        Constraints.Add(maxContentWidth);
                        maxContentHeight = ls.AddConstraint(-1, top, 1, bottom, OperatorType.LE, maxContentSize.Height);
                        Constraints.Add(maxContentHeight);
                    }
                    else
                    {
                        maxContentWidth.RightSide = maxContentSize.Width;
                        maxContentHeight.RightSide = maxContentSize.Height;
                    }
                }
                else childArea.MaxContentSize = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Preferred size of the area's content. May be different from the preferred size of the area.
        /// Manual changes of PrefContentSize are ignored unless AutoPrefContentSize is set to false.
        /// </summary>
        public Size PrefContentSize
        {
            get { return (childArea == null) ? prefContentSize : childArea.prefContentSize; }
            set
            {
                if (childArea == null)
                {
                    prefContentSize = value;
                    if (prefContentWidth == null) // no pref constraints set yet
                    {
                        prefContentWidth = ls.AddConstraint(-1, left, 1, right,
                            OperatorType.EQ, prefContentSize.Width, shrinkRigidity.Width, expandRigidity.Width);
                        Constraints.Add(prefContentWidth);
                        prefContentHeight = ls.AddConstraint(-1, top, 1, bottom,
                            OperatorType.EQ, prefContentSize.Height, shrinkRigidity.Height, expandRigidity.Height);
                        Constraints.Add(prefContentHeight);
                    }
                    else
                    {
                        prefContentWidth.RightSide = value.Width;
                        prefContentHeight.RightSide = value.Height;
                    }
                }
                else childArea.PrefContentSize = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// The reluctance with which the area's content shrinks below its preferred size.
        /// The bigger the less likely is such shrinking.
        /// </summary>
        public SizeF ShrinkRigidity
        {
            get { return (childArea == null) ? shrinkRigidity : childArea.shrinkRigidity; }
            set
            {
                if (childArea == null)
                {
                    shrinkRigidity = value;
                    if (prefContentWidth != null) // rigidities are only relevant if a preferred size is set
                    {
                        prefContentWidth.SetPenaltyNeg(value.Width);
                        prefContentHeight.SetPenaltyNeg(value.Height);
                    }
                }
                else childArea.ShrinkRigidity = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// The reluctance with which the area's content expands over its preferred size.
        /// The bigger the less likely is such expansion.
        /// </summary>
        public SizeF ExpandRigidity
        {
            get { return (childArea == null) ? expandRigidity : childArea.expandRigidity; }
            set
            {
                if (childArea == null)
                {
                    expandRigidity = value;
                    if (prefContentWidth != null) // rigidities are only relevant if a preferred size is set
                    {
                        prefContentWidth.SetPenaltyPos(value.Width);
                        prefContentHeight.SetPenaltyPos(value.Height);
                    }
                }
                else childArea.ExpandRigidity = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Aspect ratio of the area's content. May be different from the aspect ratio of the area.
        /// </summary>
        public double ContentAspectRatio
        {
            get { return (childArea == null) ? contentAspectRatio : childArea.contentAspectRatio; }
            set
            {
                if (childArea == null)
                {
                    contentAspectRatio = value;
                    if (contentAspectRatioC == null) // no aspect ratio constraint set yet
                    {
                        contentAspectRatioC = ls.AddConstraint(-1, left, 1, right, value, top, -value, bottom, 
                            OperatorType.EQ, 0);
                        Constraints.Add(contentAspectRatioC);
                    }
                    else
                    {
                        contentAspectRatioC.SetLeftSide(-1, left, 1, right, value, top, -value, bottom);
                    }
                }
                else childArea.ContentAspectRatio = value;
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Horizontal alignment of the content in its area.
        /// </summary>
        public HorizontalAlignment HAlignment
        {
            get { return hAlignment; }
            set
            {
                hAlignment = value;
                updateHorizontal();
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Vertical alignment of the content in its area.
        /// </summary>
        public VerticalAlignment VAlignment
        {
            get { return vAlignment; }
            set
            {
                vAlignment = value;
                updateVertical();
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Left margin between area and its content.
        /// </summary>
        public int LeftMargin
        {
            get { return leftMargin; }
            set
            {
                leftMargin = value;
                updateHorizontal();
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Top margin between area and its content.
        /// </summary>
        public int TopMargin
        {
            get { return topMargin; }
            set
            {
                topMargin = value;
                updateVertical();
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Right margin between area and its content.
        /// </summary>
        public int RightMargin
        {
            get { return rightMargin; }
            set
            {
                rightMargin = value;
                updateHorizontal();
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Bottom margin between area and its content.
        /// </summary>
        public int BottomMargin
        {
            get { return bottomMargin; }
            set
            {
                bottomMargin = value;
                updateVertical();
                ls.InvalidateLayout();
            }
        }

        /// <summary>
        /// Adds a childArea to this area, together with constraints that specify the relative location of the 
        /// childArea within this area. It is called when such a childArea becomes necessary, 
        /// i.e. when the user requests margins or special alignment.
        /// </summary>
        void initChildArea()
        {
            // add a child area with new tabs,
            // and add constraints that set its tabs to be equal to the 
            // coresponding tabs of this area (for a start)
            childArea = new Area(ls, new XTab(ls), new YTab(ls), new XTab(ls), new YTab(ls), content, new Size(0, 0));
            leftConstraint = Left.IsEqual(childArea.Left);
            Constraints.Add(leftConstraint);
            topConstraint = Top.IsEqual(childArea.Top);
            Constraints.Add(topConstraint);
            rightConstraint = childArea.Right.IsEqual(Right);
            Constraints.Add(rightConstraint);
            bottomConstraint = childArea.Bottom.IsEqual(Bottom);
            Constraints.Add(bottomConstraint);

            // remove the minimum content size constraints from this area
            // and copy the minimum content size setting to the childArea
            Constraints.Remove(minContentWidth);
            minContentWidth.Remove();
            minContentWidth = childArea.minContentWidth;
            Constraints.Remove(minContentHeight);
            minContentHeight.Remove();
            minContentHeight = childArea.minContentHeight;
            childArea.MinContentSize = minContentSize;

            // if there are maximum content size constraints on this area, 
            // change them so that they refer to the tabs of the childArea 
            // and copy the minimum content size settings to the childArea
            if (maxContentWidth != null)
            {
                childArea.maxContentSize = maxContentSize;
                childArea.maxContentWidth = maxContentWidth;
                maxContentWidth.SetLeftSide(-1, childArea.Left, 1, childArea.Right);
                childArea.maxContentHeight = maxContentHeight;
                maxContentHeight.SetLeftSide(-1, childArea.Top, 1, childArea.Bottom);
            }

            // if there are preferred content size constraints on this area, 
            // change them so that they refer to the tabs of the childArea 
            // and copy the preferred content size settings to the childArea
            if (prefContentHeight != null)
            {
                childArea.prefContentSize = prefContentSize;
                childArea.shrinkRigidity = shrinkRigidity;
                childArea.expandRigidity = expandRigidity;
                childArea.prefContentWidth = prefContentWidth;
                prefContentWidth.SetLeftSide(-1, childArea.Left, 1, childArea.Right);
                childArea.prefContentHeight = prefContentHeight;
                prefContentHeight.SetLeftSide(-1, childArea.Top, 1, childArea.Bottom);
            }
        }

        /// <summary>
        /// Update the constraints for horizontal margins and alignment.
        /// </summary>
        void updateHorizontal()
        {
            // if the area does not have a childAdrea yet, this is the time to add it
            if (childArea == null) initChildArea();

            // change the constraints leftConstraint and rightConstraint so that the horizontal alignment 
            // and margins of the childArea within this area are as specified by the user
            if (hAlignment == HorizontalAlignment.LEFT)
            {
                leftConstraint.SetLeftSide(-1, Left, 1, childArea.Left);
                leftConstraint.Op = OperatorType.EQ;
                leftConstraint.RightSide = leftMargin;
                rightConstraint.SetLeftSide(-1, childArea.Right, 1, Right);
                rightConstraint.Op = OperatorType.GE;
                rightConstraint.RightSide = rightMargin;
            }
            else if (hAlignment == HorizontalAlignment.RIGHT)
            {
                leftConstraint.SetLeftSide(-1, Left, 1, childArea.Left);
                leftConstraint.Op = OperatorType.GE;
                leftConstraint.RightSide = leftMargin;
                rightConstraint.SetLeftSide(-1, childArea.Right, 1, Right);
                rightConstraint.Op = OperatorType.EQ;
                rightConstraint.RightSide = rightMargin;
            }
            else if (hAlignment == HorizontalAlignment.CENTER)
            {
                leftConstraint.SetLeftSide(-1, Left, 1, childArea.Left);
                leftConstraint.Op = OperatorType.GE;
                leftConstraint.RightSide = Math.Max(leftMargin, rightMargin);
                rightConstraint.SetLeftSide(-1, Left, 1, childArea.Left, 1, childArea.Right, -1, Right);               
                rightConstraint.Op = OperatorType.EQ;
                rightConstraint.RightSide = 0;
            }
            else if (hAlignment == HorizontalAlignment.FILL)
            {
                leftConstraint.SetLeftSide(-1, Left, 1, childArea.Left);
                leftConstraint.Op = OperatorType.EQ;
                leftConstraint.RightSide = leftMargin;
                rightConstraint.SetLeftSide(-1, childArea.Right, 1, Right);
                rightConstraint.Op = OperatorType.EQ;
                rightConstraint.RightSide = rightMargin;
            }
            else if (hAlignment == HorizontalAlignment.NONE)
            {
                leftConstraint.SetLeftSide(-1, Left, 1, childArea.Left);
                leftConstraint.Op = OperatorType.GE;
                leftConstraint.RightSide = leftMargin;
                rightConstraint.SetLeftSide(-1, childArea.Right, 1, Right);
                rightConstraint.Op = OperatorType.GE;
                rightConstraint.RightSide = rightMargin;
            }
        }

        /// <summary>
        /// Update the constraints for vertical margins and alignment.
        /// </summary>
        void updateVertical()
        {
            // if the area does not have a childAdrea yet, this is the time to add it
            if (childArea == null) initChildArea();

            // change the constraints topConstraint and bottomConstraint so that the vertical alignment 
            // and margins of the childArea within this area are as specified by the user
            if (vAlignment == VerticalAlignment.TOP)
            {
                topConstraint.SetLeftSide(-1, Top, 1, childArea.Top);
                topConstraint.Op = OperatorType.EQ;
                topConstraint.RightSide = topMargin;
                bottomConstraint.SetLeftSide(-1, childArea.Bottom,1,Bottom);
                bottomConstraint.Op = OperatorType.GE;
                bottomConstraint.RightSide = bottomMargin;
            }
            else if (vAlignment == VerticalAlignment.BOTTOM)
            {
                topConstraint.SetLeftSide(-1, Top, 1, childArea.Top);
                topConstraint.Op = OperatorType.GE;
                topConstraint.RightSide = topMargin;
                bottomConstraint.SetLeftSide(-1, childArea.Bottom, 1, Bottom);
                bottomConstraint.Op = OperatorType.EQ;
                bottomConstraint.RightSide = bottomMargin;
            }
            else if (vAlignment == VerticalAlignment.CENTER)
            {
                topConstraint.SetLeftSide(-1, Top, 1, childArea.Top);
                topConstraint.Op = OperatorType.GE;
                topConstraint.RightSide = Math.Max(topMargin, bottomMargin);
                bottomConstraint.SetLeftSide(-1, Top, 1, childArea.Top, 1, childArea.Bottom, -1, Bottom);     
                bottomConstraint.Op = OperatorType.EQ;
                bottomConstraint.RightSide = 0;
            }
            else if (vAlignment == VerticalAlignment.FILL)
            {
                topConstraint.SetLeftSide(-1, Top, 1, childArea.Top);
                topConstraint.Op = OperatorType.EQ;
                topConstraint.RightSide = topMargin;
                bottomConstraint.SetLeftSide(-1, childArea.Bottom, 1, Bottom);
                bottomConstraint.Op = OperatorType.EQ;
                bottomConstraint.RightSide = bottomMargin;
            }
            else if (vAlignment == VerticalAlignment.NONE)
            {
                topConstraint.SetLeftSide(-1, Top, 1, childArea.Top);
                topConstraint.Op = OperatorType.GE;
                topConstraint.RightSide = topMargin;
                bottomConstraint.SetLeftSide(-1, childArea.Bottom, 1, Bottom);
                bottomConstraint.Op = OperatorType.GE;
                bottomConstraint.RightSide = bottomMargin;
            }
        }

        /// <summary>
        /// Sets the preferred size according to the content's PreferredSize method, 
        /// and the rigidities according to heuristics.
        /// </summary>
        public void SetDefaultPrefContentSize()
        {
            if (Content == null)
            {
                /*
                PrefContentSize = new Size(0, 0);
                ShrinkRigidity = new SizeF(0, 0);
                ExpandRigidity = new SizeF(0, 0);
                 */
                return;
            }

            if (PrefContentSize != Content.PreferredSize)
            {
                PrefContentSize = Content.PreferredSize;
                ls.InvalidateLayout();
            }

            /* use heuristics for rigidities:
             * controls with naturally constant content size 
             * are less likely to change their size */
            if (Content is Button
                || Content is RadioButton
                || Content is CheckBox
                || Content is Label
                || Content is LinkLabel
                || Content is PictureBox
                || Content is ProgressBar
                || Content is NumericUpDown)
            {
                ShrinkRigidity = new SizeF(4, 4);
                ExpandRigidity = new SizeF(3, 3);
            }
            else
            {
                ShrinkRigidity = new SizeF(2, 2);
                ExpandRigidity = new SizeF(1, 1);
            }
        }

        internal void DoLayout(bool editing)
        {
            if (Content != null) // empty areas need no layout
            { 
                // if there is a childArea, then it is the childArea that actually contains the content
                Area a = (childArea != null) ? childArea : this;

                // set content location and size
                a.Content.Location = new Point((int)Math.Round(a.Left.Value), (int)Math.Round(a.Top.Value));
                a.Content.Width = (int)Math.Round(a.Right.Value - a.Left.Value);
                a.Content.Height = (int)Math.Round(a.Bottom.Value - a.Top.Value);
            }
            /*
            if (editing)
            {
                AreaControl.Location = new Point((int)Math.Round(Left.Value), (int)Math.Round(Top.Value));
                AreaControl.Width = (int)Math.Round(Right.Value - Left.Value);
                AreaControl.Height = (int)Math.Round(Bottom.Value - Top.Value);
                AreaControl.Invalidate();
             
            }
        */
        }

        public override string ToString()
        {
            return "Area(" + left.ToString() + "," + top.ToString() + ","
                + right.ToString() + "," + bottom.ToString() + ")";
        }

        /// <summary>
        /// Sets the width of the area to be the same as the width of the given area.
        /// </summary>
        /// <param name="a">the area that should have the same width</param>
        /// <returns>the same-width constraint</returns>
        public Constraint HasSameWidthAs(Area a)
        {
            return ls.AddConstraint(-1,left ,1,right, 1,a.left, -1,a.right,               
                OperatorType.EQ, 0);
        }

        /// <summary>
        /// Sets the height of the area to be the same as the height of the given area.
        /// </summary>
        /// <param name="a">the area that should have the same height</param>
        /// <returns>the same-height constraint</returns>
        public Constraint HasSameHeightAs(Area a)
        {
            return ls.AddConstraint(-1,top ,1,bottom, 1,a.top, -1,a.bottom,    
                OperatorType.EQ, 0);
        }

        /// <summary>
        /// Sets the size of the area to be the same as the size of the given area.
        /// </summary>
        /// <param name="a">the area that should have the same size</param>
        /// <returns>a list containing a same-width and a same-height constraint</returns>
        public List<Constraint> HasSameSizetAs(Area a)
        {
            List<Constraint> l = new List<Constraint>();
            l.Add(this.HasSameWidthAs(a));
            l.Add(this.HasSameHeightAs(a));
            return l;
        }

        internal Area(LayoutSpec ls, XTab left, YTab top, XTab right, YTab bottom, Control content,
            Size minContentSize)
        {
            this.ls = ls;
            this.left = left;
            this.right = right;
            this.top = top;
            this.bottom = bottom;
            Content = content;
            this.minContentSize = minContentSize;

            // adds the two essential constraints of the area that make sure that the left x-tab is really
            // to the left of the right x-tab, and the top y-tab really above the bottom y-tab
            minContentWidth = ls.AddConstraint(-1,left, 1, right, OperatorType.GE, minContentSize.Width);
            Constraints.Add(minContentWidth);
            minContentHeight = ls.AddConstraint(-1, top, 1, bottom, OperatorType.GE, minContentSize.Height);
            Constraints.Add(minContentHeight);
        }

        internal Area(LayoutSpec ls, Row row, Column column, Control content, Size minContentSize)
            : this(ls, column.Left, row.Top, column.Right, row.Bottom, content, minContentSize)
        {
            this.row = row;
            this.column = column;
        }

        /// <summary>
        /// Removes the area from its specification.
        /// </summary>
        public void Remove()
        {
            if(childArea != null) childArea.Remove();
            foreach (Constraint c in Constraints) c.Remove();
            ls.Areas.Remove(this);
        }
    }
}
