using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Windows.Forms.Layout;
using LinearProgramming;
using System.IO;

namespace ALM
{
    /// <summary>
    /// Vertical grid line (x-tab).
    /// </summary>
    public class XTab : Variable
    {
        internal XTab(LayoutSpec ls) : base(ls) { }

        /// <summary>
        /// Property signifying if there is a constraint which relates
        /// this tab to a different tab that is further to the left.
        /// Only used for reverse engineering.
        /// </summary>
        internal bool LeftLink = false;
    }

    /// <summary>
    /// Horizontal grid line (y-tab).
    /// </summary>
    public class YTab : Variable
    {
        internal YTab(LayoutSpec ls) : base(ls) { }

        /// <summary>
        /// Property signifying if there is a constraint which relates
        /// this tab to a different tab that is further to the top.
        /// Only used for reverse engineering.
        /// </summary>
        internal bool TopLink = false;
    }

    /// <summary>
    /// Layout specification.
    /// </summary>
    public class LayoutSpec : LinearSpec
    {
        /// <summary>
        /// The areas that were added to the specification.
        /// </summary>
        public List<Area> Areas = new List<Area>();

        /// <summary>
        /// The rows that were added to the specification.
        /// </summary>
        public List<Row> Rows = new List<Row>();

        /// <summary>
        /// The columns that were added to the specification.
        /// </summary>
        public List<Column> Columns = new List<Column>();

        /// <summary>
        /// X-tab for the left border of the GUI.
        /// </summary>
        public XTab Left;

        /// <summary>
        /// X-tab for the right border of the GUI.
        /// </summary>
        public XTab Right;

        /// <summary>
        /// Y-tab for the top border of the GUI.
        /// </summary>
        public YTab Top;

        /// <summary>
        /// Y-tab for the bottom border of the GUI.
        /// </summary>
        public YTab Bottom;

        public LayoutSpec()
            : base()
        {
            Left = new XTab(this);
            Right = new XTab(this);
            Top = new YTab(this);
            Bottom = new YTab(this);

            // the Left tab is always at x-position 0, and the Top tab is always at y-position 0
            Left.SetRange(0, 0);
            Top.SetRange(0, 0);
        }

        public void SolveLayout()
        {
            // if AutoPrefContentSize is set on an area, readjust its PrefContentSize and rigidity settings
            foreach (Area a in Areas)
                if (a.AutoPrefContentSize) a.SetDefaultPrefContentSize();

            // try to solve the layout until the result is OPTIMAL or INFEASIBLE, maximally 15 tries
            // sometimes the solving algorithm encounters numerical problems (NUMFAILURE), and
            // repeating the solving often helps to overcome them
            StreamWriter sw = new StreamWriter("performance.txt", true);
            ResultType result = ResultType.ERROR;
            for (int tries = 0; tries < 15; tries++)
            {
                result = Solve();
                sw.WriteLine(result + "\t" + SolvingTime + "ms"
                    + "\t#vars=" + Variables.Count + "\t#constraints=" + Constraints.Count);
                if (result == ResultType.OPTIMAL || result == ResultType.INFEASIBLE) break;
            }
            sw.Close();
        }

        // cached layout values
        // need to be invalidated whenever the layout specification is changed
        Size minSize = Area.UNDEFINED_SIZE;
        Size maxSize = Area.UNDEFINED_SIZE;
        Size preferredSize = Area.UNDEFINED_SIZE;

        public void InvalidateLayout()
        {
            minSize = Area.UNDEFINED_SIZE;
            maxSize = Area.UNDEFINED_SIZE;
            preferredSize = Area.UNDEFINED_SIZE;
        }

        public Size MinSize
        {
            get
            {
                if (minSize == Area.UNDEFINED_SIZE)
                    minSize = CalculateMinSize().ToSize();
                return minSize;
            }
        }

        public Size MaxSize
        {
            get
            {
                if (maxSize == Area.UNDEFINED_SIZE)
                    maxSize = CalculateMaxSize().ToSize();
                return maxSize;
            }
        }

        public Size PreferredSize
        {
            get
            {
                if (preferredSize == Area.UNDEFINED_SIZE)
                    preferredSize = CalculatePreferredSize().ToSize();
                return preferredSize;
            }
        }

        SizeF CalculateMinSize()
        {
            List<Summand> buf = ObjFunctionSummands;
            ObjFunctionSummands = new List<Summand>();
            AddObjFunctionSummand(1, Right);
            AddObjFunctionSummand(1, Bottom);
            SolveLayout();
            ObjFunctionSummands = buf;
            UpdateObjFunction();

            if (Result == ResultType.UNBOUNDED)
                return Area.MIN_SIZE;
            if (Result != ResultType.OPTIMAL)
            {
                Save("failed-layout.txt");
                throw new Exception("Could not solve the layout specification ("
                    + Result.ToString() + "). Saved specification in file failed-layout.txt");
            }
            return new SizeF(
                    (float)(Right.Value - Left.Value),
                    (float)(Bottom.Value - Top.Value));
        }

        SizeF CalculateMaxSize()
        {
            List<Summand> buf = ObjFunctionSummands;
            ObjFunctionSummands = new List<Summand>();
            AddObjFunctionSummand(-1, Right);
            AddObjFunctionSummand(-1, Bottom);
            SolveLayout();
            ObjFunctionSummands = buf;
            UpdateObjFunction();

            if (Result == ResultType.UNBOUNDED)
                return new SizeF(999999, 999999); // workaround. should be return Area.MAX_SIZE
            if (Result != ResultType.OPTIMAL)
            {
                Save("failed-layout.txt");
                throw new Exception("Could not solve the layout specification ("
                    + Result.ToString() + "). Saved specification in file failed-layout.txt");
            }
            return new SizeF(
                    (float)(Right.Value - Left.Value),
                    (float)(Bottom.Value - Top.Value));
        }

        SizeF CalculatePreferredSize()
        {
            SolveLayout();
            if (Result != ResultType.OPTIMAL)
            {
                Save("failed-layout.txt");
                throw new Exception("Could not solve the layout specification ("
                    + Result.ToString() + "). Saved specification in file failed-layout.txt");
            }
            return new SizeF(
                (float)(Right.Value - Left.Value),
                (float)(Bottom.Value - Top.Value));
        }

        public bool HasHeightForWidth { get { return true; } }

        public void GetHeightForWidth(float width, out float min, out float max, out float preferred)
        {
            double rightMin = Right.Min;
            double rightMax = Right.Max;
            Right.SetRange(width, width);

            min = CalculateMinSize().Height;
            max = CalculateMaxSize().Height;
            preferred = CalculatePreferredSize().Height;
            Right.SetRange(rightMin, rightMax);
        }

        /// <summary>
        /// Adds a new x-tab to the specification.
        /// </summary>
        /// <returns>the new x-tab</returns>
        public XTab AddXTab()
        {
            return new XTab(this);
        }

        /// <summary>
        /// Adds a new y-tab to the specification.
        /// </summary>
        /// <returns>the new y-tab</returns>
        public YTab AddYTab()
        {
            return new YTab(this);
        }

        /// <summary>
        /// Adds a new row to the specification.
        /// </summary>
        /// <returns>the new row</returns>
        public Row AddRow()
        {
            Row row = new Row(this);
            Rows.Add(row);
            return row;
        }

        /// <summary>
        /// Adds a new column to the specification.
        /// </summary>
        /// <returns>the new column</returns>
        public Column AddColumn()
        {
            Column column = new Column(this);
            Columns.Add(column);
            return column;
        }

        /// <summary>
        /// Adds a new row to the specification that is glued to the given y-tabs.
        /// </summary>
        /// <returns>the new row</returns>
        public Row AddRow(YTab top, YTab bottom)
        {
            Row row = new Row(this);
            if (top != null) row.Constraints.Add(row.Top.IsEqual(top));
            if (bottom != null) row.Constraints.Add(row.Bottom.IsEqual(bottom));
            Rows.Add(row);
            return row;
        }

        /// <summary>
        /// Adds a new column to the specification that is glued to the given x-tabs.
        /// </summary>
        /// <returns>the new column</returns>
        public Column AddColumn(XTab left, XTab right)
        {
            Column column = new Column(this);
            if (left != null) column.Constraints.Add(column.Left.IsEqual(left));
            if (right != null) column.Constraints.Add(column.Right.IsEqual(right));
            Columns.Add(column);
            return column;
        }

        /// <summary>
        /// Adds a new area to the specification, setting only the necessary minimum size constraints.
        /// </summary>
        /// <param name="left">left border</param>
        /// <param name="top">top border</param>
        /// <param name="right">right border</param>
        /// <param name="bottom">bottom border</param>
        /// <param name="content">the control which is the area content</param>
        /// <param name="minContentSize">minimum content size</param>
        /// <returns>the new area</returns>
        public Area AddArea(XTab left, YTab top, XTab right, YTab bottom, Control content,
            Size minContentSize)
        {
            InvalidateLayout();
            Area a = new Area(this, left, top, right, bottom, content, minContentSize);
            Areas.Add(a);
            return a;
        }

        /// <summary>
        /// Adds a new area to the specification, setting only the necessary minimum size constraints.
        /// </summary>
        /// <param name="row">the row that defines the top and bottom border</param>
        /// <param name="column">to column that defines the left and right border</param>
        /// <param name="content">the control which is the area content</param>
        /// <param name="minContentSize">minimum content size</param>
        /// <returns>the new area</returns>
        public Area AddArea(Row row, Column column, Control content,
            Size minContentSize)
        {
            InvalidateLayout();
            Area a = new Area(this, row, column, content, minContentSize);
            Areas.Add(a);
            return a;
        }
        
        /// <summary>
        /// Adds a new area to the specification, automatically setting preferred size constraints.
        /// </summary>
        /// <param name="left">left border</param>
        /// <param name="top">top border</param>
        /// <param name="right">right border</param>
        /// <param name="bottom">bottom border</param>
        /// <param name="content">the control which is the area content</param>
        /// <returns>the new area</returns>
        public Area AddArea(XTab left, YTab top, XTab right, YTab bottom, Control content)
        {
            InvalidateLayout();
            Area a = new Area(this, left, top, right, bottom, content, new Size(0, 0));
            a.SetDefaultPrefContentSize();
            a.AutoPrefContentSize = false;
            Areas.Add(a);
            return a;
        }
        
        /// <summary>
        /// Adds a new area to the specification, automatically setting preferred size constraints.
        /// </summary>
        /// <param name="row">the row that defines the top and bottom border</param>
        /// <param name="column">to column that defines the left and right border</param>
        /// <param name="content">the control which is the area content</param>
        /// <returns>the new area</returns>
        public Area AddArea(Row row, Column column, Control content)
        {
            InvalidateLayout();
            Area a = new Area(this, row, column, content, new Size(0, 0));
            a.SetDefaultPrefContentSize();
            a.AutoPrefContentSize = false;
            Areas.Add(a);
            return a;
        }

        /// <summary>
        /// Finds the area that contains the given control.
        /// </summary>
        /// <param name="c">the control to look for</param>
        /// <returns>the area that contains the control</returns>
        public Area AreaOf(Control c)
        {
            foreach (Area a in Areas)
                if (a.Content == c) return a;
            return null;
        }
    }
}
