using System;
using System.Collections.Generic;
using System.Text;
using LinearProgramming;

namespace ALM
{
    /// <summary>
    /// Represents a row defined by two y-tabs.
    /// </summary>
    public class Row
    {
        /// <summary>
        /// The layout specification this row belongs to. 
        /// </summary>
        internal LayoutSpec ls;

        internal YTab top, bottom;
        Row previous, next;
        
        /// <summary>
        /// The constraint that fixes the top tab of this row to the bottom tab of the previous row.
        /// </summary>
        Constraint previousGlue;

        /// <summary>
        /// The constraint that fixes the bottom tab of this row to the top tab of the next row.
        /// </summary>
        Constraint nextGlue;

        /// <summary>
        /// The top boundary of the row.
        /// </summary>
        public YTab Top { get { return top; } }

        /// <summary>
        /// The bottom boundary of the row.
        /// </summary>
        public YTab Bottom { get { return bottom; } }

        /// <summary>
        /// Constraints that are removed when the row is removed.
        /// </summary>
        public List<Constraint> Constraints = new List<Constraint>();

        /// <summary>
        /// The row directly above this row. May be null.
        /// </summary>
        public Row Previous
        {
            get { return previous; }
            set
            {
                // if there should be no row directly above this row, then we have to
                // separate any such row and can remove any constraint that was used 
                // to glue this row to it
                if (value == null)
                {
                    if (previous == null) return;
                    previous.next = null;
                    previous.nextGlue = null;
                    previous = null;
                    previousGlue.Remove();
                    previousGlue = null;
                    return;
                }

                // otherwise we have to set up the pointers and the glue constraint accordingly
                if (value.next != null) value.Next = null;
                if (previous != null) Previous = null;

                previous = value;
                previous.next = this;
                previousGlue = value.nextGlue = value.bottom.IsEqual(top);
            }
        }

        /// <summary>
        /// The row directly below this row. May be null.
        /// </summary>
        public Row Next
        {
            get { return next; }
            set
            {
                // if there should be no row directly below this row, then we have to
                // separate any such row and can remove any constraint that was used 
                // to glue this row to it
                if (value == null)
                {
                    if (next == null) return;
                    next.previous = null;
                    next.previousGlue = null;
                    next = null;
                    nextGlue.Remove();
                    nextGlue = null;
                    return;
                }

                // otherwise we have to set up the pointers and the glue constraint accordingly
                if (value.previous != null) value.Previous = null;
                if (next != null) Next = null;

                next = value;
                next.previous = this;
                nextGlue = value.previousGlue = bottom.IsEqual(value.top);
            }
        }

        public override string ToString()
        {
            return "Row(" + top.ToString() + ", " + bottom.ToString() + ")";
        }

        /// <summary>
        /// Inserts the given row directly above this row.
        /// </summary>
        /// <param name="r">the row to insert</param>
        public void InsertBefore(Row r)
        {
            Previous = r.previous;
            Next = r;
        }

        /// <summary>
        /// Inserts the given row directly below this row.
        /// </summary>
        /// <param name="r">the row to insert</param>
        public void InsertAfter(Row r)
        {
            Next = r.next;
            Previous = r;
        }

        /// <summary>
        /// Constrains this row to have the same height as the given row.
        /// </summary>
        /// <param name="row">The row that should have the same height</param>
        /// <returns>the resulting same-height constraint</returns>
        public Constraint HasSameHeightAs(Row row)
        {
            Constraint c = ls.AddConstraint(
                -1, top, 1, bottom, 1, row.top, -1, row.bottom, OperatorType.EQ, 0);
            Constraints.Add(c);
            return c;
        }

        internal Row(LayoutSpec ls)
        {
            this.ls = ls;
            top = new YTab(ls);
            bottom = new YTab(ls);
        }

        public void Unlink()
        {
            if (previous != null) previous.Next = next;
            else if (next != null) next.Previous = previous;
        }

        /// <summary>
        /// Removes the row from the specification.
        /// </summary>
        public void Remove()
        {
            if (previous != null) previous.Next = next;
            foreach (Constraint c in Constraints) c.Remove();
            top.Remove();
            bottom.Remove();
        }
    }

    /// <summary>
    /// Represents a column defined by two x-tabs.
    /// </summary>
    public class Column
    {
        /// <summary>
        /// The layout specification this column belongs to. 
        /// </summary>
        internal LayoutSpec ls;

        internal XTab left, right;
        Column previous, next;

        /// <summary>
        /// The constraint that fixes the left tab of this column to the right tab of the previous column.
        /// </summary>
        Constraint previousGlue;

        /// <summary>
        /// The constraint that fixes the right tab of this column to the left tab of the next column.
        /// </summary>
        Constraint nextGlue;

        /// <summary>
        /// The left boundary of the column.
        /// </summary>
        public XTab Left { get { return left; } }

        /// <summary>
        /// The right boundary of the column.
        /// </summary>
        public XTab Right { get { return right; } }

        /// <summary>
        /// Constraints that are removed when the column is removed.
        /// </summary>
        public List<Constraint> Constraints = new List<Constraint>();

        /// <summary>
        /// The column directly to the left of this column. May be null.
        /// </summary>
        public Column Previous
        {
            get { return previous; }
            set
            {
                // if there should be no column directly left of this column, then we have to
                // separate any such column and can remove any constraint that was used 
                // to glue this column to it
                if (value == null)
                {
                    if (previous == null) return;
                    previous.next = null;
                    previous.nextGlue = null;
                    previous = null;
                    previousGlue.Remove();
                    previousGlue = null;
                    return;
                }

                // otherwise we have to set up the pointers and the glue constraint accordingly
                if (value.next != null) value.Next = null;
                if (previous != null) Previous = null;

                previous = value;
                previous.next = this;
                previousGlue = value.nextGlue = value.right.IsEqual(left);
            }
        }

        /// <summary>
        /// The column directly to the right of this column. May be null.
        /// </summary>
        public Column Next
        {
            get { return next; }
            set
            {
                // if there should be no column directly right of this column, then we have to
                // separate any such column and can remove any constraint that was used 
                // to glue this column to it
                if (value == null)
                {
                    if (next == null) return;
                    next.previous = null;
                    next.previousGlue = null;
                    next = null;
                    nextGlue.Remove();
                    nextGlue = null;
                    return;
                }

                // otherwise we have to set up the pointers and the glue constraint accordingly
                if (value.previous != null) value.Previous = null;
                if (next != null) Next = null;

                next = value;
                next.previous = this;
                nextGlue = value.previousGlue = right.IsEqual(value.left);
            }
        }

        public override string ToString()
        {
            return "Column(" + left.ToString() + ", " + right.ToString() + ")";
        }

        /// <summary>
        /// Inserts the given column directly to the left of this column.
        /// </summary>
        /// <param name="c">the column to insert</param>
        public void InsertBefore(Column c)
        {
            Previous = c.previous;
            Next = c;
        }

        /// <summary>
        /// Inserts the given column directly to the right of this column.
        /// </summary>
        /// <param name="c">the column to insert</param>
        public void InsertAfter(Column c)
        {
            Next = c.next;
            Previous = c;
        }

        /// <summary>
        /// Constrains this column to have the same width as the given column.
        /// </summary>
        /// <param name="column">The column that should have the same width</param>
        /// <returns>the resulting same-width constraint</returns>
        public Constraint HasSameWidthAs(Column column)
        {
            Constraint c = ls.AddConstraint(
                -1, left, 1, right, 1, column.left, -1, column.right, OperatorType.EQ, 0);
            Constraints.Add(c);
            return c;
        }

        internal Column(LayoutSpec ls)
        {
            this.ls = ls;
            left = new XTab(ls);
            right = new XTab(ls);
        }

        public void Unlink()
        {
            if (previous != null) previous.Next = next;
            else if (next != null) next.Previous = previous;
        }

        /// <summary>
        /// Removes the column from the specification.
        /// </summary>
        public void Remove()
        {
            if (previous != null) previous.Next = next;
            foreach (Constraint c in Constraints) c.Remove();
            left.Remove();
            right.Remove();
        }
    }
}
